package org.simantics.databoard.channel;

import org.simantics.databoard.Accessors;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.error.RuntimeAccessorConstructionException;
import org.simantics.databoard.accessor.error.RuntimeAccessorException;
import org.simantics.databoard.binding.Binding;

/**
 * Service exception is thrown by Service Handler when there was problem in 
 * processing the service request. 
 *
 * @author Toni Kalajainen <toni.kalajainen@iki.fi>
 */
public abstract class ServiceException extends Exception {

	private static final long serialVersionUID = 1L;

	Accessor error;
	
	public ServiceException(Accessor error) {
		this.error = error;
	}
	
	public ServiceException(Binding binding, Object value) {
		try {
			this.error = Accessors.getAccessor(binding, value);
		} catch (AccessorConstructionException e) {
			throw new RuntimeAccessorConstructionException(e);
		}
	}
	
	public Accessor getError() {
		return error;
	}
	
	public Object getErrorType() {
		return error.type();
	}
	
	public Object getError(Binding binding) {
		try {
			return error.getValue(binding);
		} catch (AccessorException e) {
			throw new RuntimeAccessorException(e);
		}
	}
	
}
